/*
   M5SScope
   --------
   v 0.80 2018/05/17 Bata
   (c) 2018- Goji(goji2100@gmail.com)
*/

#define M5SScope_VERSION  "V0.80"

#include <M5Stack.h>
#include "ICONs.h"
#include "font8x12.h"

#define USE_WAVE_DAC
#define FUCTION_KEY
#define DUMP_SCREEN

#define ADC_PIN0        35
#define ADC_PIN1        36
#define TEST_PIN0        2
#define TEST_PIN1        5

#ifdef  FUCTION_KEY
#define TRIG_INFO_TOP    0
#define SCOPE_HOME_XS   16    // Wave window Left
#define SCOPE_HOME_YS   14    // Wave window Top
#else
#define TRIG_INFO_TOP    2
#define SCOPE_HOME_XS   16    // Wave window Left
#define SCOPE_HOME_YS   16    // Wave window Top
#endif

#define SCOPE_WIDTH_S    (80 * 3)
#define SCOPE_HEIGHT_S   (80 * 2)

#define SCOPE_HOME_XL    0
#define SCOPE_HOME_YL    0
#define SCOPE_WIDTH_L    (80 * 4)
#define SCOPE_HEIGHT_L   (80 * 3)

#define TFT_MIN_X        0
#define TFT_MIN_Y        0
#define TFT_MAX_X      320
#define TFT_MAX_Y      240

uint16_t VW_xb = SCOPE_HOME_XS;
uint16_t VW_yb = SCOPE_HOME_YS;
uint16_t VW_xh = SCOPE_WIDTH_S;
uint16_t VW_yh = SCOPE_HEIGHT_S;

#define NUMBUFF (SCOPE_WIDTH_S * 2)
volatile uint16_t wavebuf[2][NUMBUFF];
uint8_t waverec[2][2][SCOPE_WIDTH_L];

#define DAC_VALS       100
uint8_t DACvals[DAC_VALS];

#ifdef DUMP_SCREEN
uint16_t drawBMP = 1;
uint8_t  bmpBuff[TFT_MAX_Y][TFT_MAX_X];
#else
uint16_t drawBMP = 0;
uint16_t bmpBuff[2][2]; // dummy(^^;
#endif

#define __RGB(r, g, b)  ((r << 11)|(g << 5)|b)
#define rgbBLACK     __RGB( 0,  0,  0)
#define rgbBLUE      __RGB( 0,  0, 31)
#define rgbGREEN     __RGB( 0, 63,  0)
#define rgbCYAN      __RGB( 0, 63, 31)
#define rgbRED       __RGB(31,  0,  0)
#define rgbMAGENTA   __RGB(31,  0, 31)
#define rgbYELLOW    __RGB(31, 63,  0)
#define rgbWHITE     __RGB(31, 63, 31)
#define rgbGRAY1     __RGB(20, 40, 20)
#define rgbGRAY2     __RGB(26, 52, 26)
#define rgbBLACK1    __RGB( 0,  0,  1)

uint16_t Color_tbl[] = { rgbBLACK, rgbBLUE, rgbGREEN, rgbCYAN, rgbRED, rgbMAGENTA, rgbYELLOW, rgbWHITE, rgbGRAY1, rgbGRAY2 };
typedef enum           {  ixBLACK,  ixBLUE,  ixGREEN,  ixCYAN,  ixRED,  ixMAGENTA,  ixYELLOW,  ixWHITE,  ixGRAY1,  ixGRAY2, ixBLACK1 } ixcolor_t;

const uint16_t ch1_color = ixYELLOW;
const uint16_t ch2_color = ixRED;

/* --------------------------------------------------------*/
#define	SV_CH1			    (1 << 1)
#define	SV_CH2			    (1 << 2)

#define	SV_CH_DISBL		  0
#define	SV_CH_ENABL		  1
#define	SV_CH_AC  		  0
#define	SV_CH_DC	  	  1
#define	SV_CH_X1		    0
#define	SV_CH_X10		    1

#define	SV_TR_TYPE_FREE	0
#define	SV_TR_TYPE_UP	  1
#define	SV_TR_TYPE_DWN	2
#define	SV_TR_TYPE_HI	  3
#define	SV_TR_TYPE_LOW	4

#define SV_WF_SEL_NONE  0
#ifdef USE_WAVE_DAC
#define SV_WF_SEL_RECT  1
#define SV_WF_SEL_STEP  2
#define SV_WF_SEL_SAW   3
#define SV_WF_SEL_SIN   4
#else
#define SV_WF_SEL_RECT  1
#endif

#define SV_DP_SEL_NONE  0
#define SV_DP_SEL_BMP   1
#define SV_DP_SEL_UART  2

#define SW_SELECT       1
#define SW_LEFT         2
#define SW_RIGHT        3

#define	VT_UNIT_MAX	7
const char  vd_sels[][6]	= { "100mV", "200mV", "500mV", "   1V", "   2V", "   5V", "  10V" };
const long  vd_seln[]		  = {     100,     200,     500,    1000,    2000,    5000,   10000 };

uint16_t sv_ch_sel		= (SV_CH1 | SV_CH2);					          // Ch1, Ch2 Enable
uint16_t sv_ch_vdiv[]	= {               3,              3 };	// Select = 1V
uint16_t sv_ch_prob[]	= {       SV_CH_X10,      SV_CH_X10 };	// Probe x10
int32_t  sv_ch_posn[]	= {           -3000,          -3000 };	// Pos = -3.0V
int16_t  sv_ch_poss[]	= {               0,              0 };	// <- Calc in drawGaid

uint16_t sv_tr_type[]	= {   SV_TR_TYPE_UP, SV_TR_TYPE_FREE };	// Trig type = (Free)
int32_t  sv_tr_levn[]	= {            1000,            1000 };	// Trig level (value)
int16_t  sv_tr_levs[]	= {               0,               0 };	// Trig level (scope) <- Calc in drawGaid

#define	SV_TR_SEL_NONE  0b0000
#define	SV_TR_SEL_CH1   0b0001
#define	SV_TR_SEL_CH2 	0b0010
#define	SV_TR_SEL_OR  	0b0011
#define	SV_TR_SEL_AND 	0b0100
uint16_t sv_tr_sel = SV_TR_SEL_CH1;
char tr_sels[][7]	= {	" Free ", " CH1  ", " CH2  ", "  OR  ", " AND  " };

#if (0)
#define  HZ_UNIT_MAX 12
#define	nn(n)		(((4 * n) / 2) - 1)		// (Fcy / 1000 / 1000) * n(ns) / 20 (samples / DIV)
char     hz_units[][6] = {	" 20us", " 50us", "100us", "200us", "500us", "  1ms" , "  2ms" , "  5ms" , " 10ms" , " 20ms" , " 50ms", "100ms"  };
uint32_t hz_unitn[]    = {     20 ,     50 ,    100 ,    200 ,    500 ,    1000 ,    2000 ,    5000 ,   10000 ,   20000 ,   50000,  100000  };
uint16_t hz_unitv[] 	 = {  nn(20),  nn(50), nn(100), nn(200), nn(500), nn(1000), nn(2000), nn(5000),       1 ,       2 ,       5,      10  };
uint32_t sv_hz_posn    = 0;
uint16_t sv_horz_unit	 = 5;           // Select 1ms/DIV
uint16_t sv_wf_sel = SV_WF_SEL_NONE;
#else

#define  HZ_UNIT_MAX 8
#define  nn(n)   (n / 20)
char     hz_units[][6] = { "500us" , "  1ms" , "  2ms" , "  5ms" ,  " 10ms" ,  " 20ms" ,  " 50ms",    "100ms"  };
int32_t  hz_unitn[]    = {     500 ,    1000 ,    2000 ,    5000 ,    10000 ,    20000 ,    50000,     100000  };
uint32_t hz_unitv[]    = { nn( 500), nn(1000), nn(2000), nn(5000), nn(10000), nn(20000), nn(50000), nn(100000) }; // us * 20 = ms/div
int32_t  sv_hz_posn    = 0;
uint16_t sv_horz_unit  = 1;           // Select 1ms/DIV
uint16_t sv_wf_sel = SV_WF_SEL_RECT;  // for DEMO!
uint16_t sv_dp_sel = SV_DP_SEL_NONE;
#endif

#define delay_us(usec) delayMicro(usec)
#define delay_ms(msec) delay(msec)

hw_timer_t *Tick1Timer = NULL;
volatile SemaphoreHandle_t Tick1Semaphore;
portMUX_TYPE Tick1TimerMux = portMUX_INITIALIZER_UNLOCKED;

volatile int waveGetRun   = 0, waveGenStart = 0;
volatile int dmaCountP, dmaCountG;

void IRAM_ATTR Tick1TimerISR() {
  portENTER_CRITICAL_ISR(&Tick1TimerMux);
  {
    if (waveGetRun) {
      if (waveGenStart == 0) {
        waveGenStart++;
        dmaCountP = dmaCountG = 0;
      }
#if (0)
      //digitalWrite(TEST_PIN1, 1);
      dmaCountP++;
      if (dmaCountP >= NUMBUFF) dmaCountP = 0;
      if (dmaCountP == dmaCountG) {
        dmaCountG++;
        if (dmaCountG >= NUMBUFF) dmaCountG = 0;
      }
      wavebuf[0][dmaCountP] = analogRead(ADC_PIN0);
      wavebuf[1][dmaCountP] = analogRead(ADC_PIN1);
      //digitalWrite(TEST_PIN1, 0);
#else
      if (dmaCountP < (NUMBUFF - 1)) {
        //digitalWrite(TEST_PIN1, 1);
        wavebuf[0][dmaCountP] = analogRead(ADC_PIN0);
        wavebuf[1][dmaCountP] = analogRead(ADC_PIN1);
        dmaCountP++;
        //digitalWrite(TEST_PIN1, 0);
      }
#endif
    } else waveGenStart = 0;
  }
  portEXIT_CRITICAL_ISR(&Tick1TimerMux);
  xSemaphoreGiveFromISR(Tick1Semaphore, NULL);
}

void setupTick1(uint32_t n) {
  if (waveGetRun == 1) waveGetRun--;
  delay(1);

  if (Tick1Timer) {
    timerEnd(Tick1Timer);
    Tick1Timer = NULL;
  }

  waveGenStart = 0;
  Tick1Semaphore = xSemaphoreCreateBinary();
  Tick1Timer = timerBegin(0,  ((n < 1000) ? 80 : 800), true);  // 1MHz / 100KHz
  timerAlarmWrite(Tick1Timer, ((n < 1000) ? n : (n / 10)), true);
  timerAttachInterrupt(Tick1Timer, &Tick1TimerISR, true);
  timerAlarmEnable(Tick1Timer);
  waveGetRun++;
}

hw_timer_t *Tick2Timer = NULL;
volatile SemaphoreHandle_t Tick2Semaphore;
portMUX_TYPE Tick2TimerMux = portMUX_INITIALIZER_UNLOCKED;

volatile int Tick2Run = 0, Tick20x = 0, Tick21x = 0, Tick20Val = 0, Tick21Val = 0;

void IRAM_ATTR Tick2TimerISR() {
  if (Tick2Run) {
#ifdef USE_WAVE_DAC
    if (++Tick20x >= 10) {
      Tick20x = 0;
      Tick20Val = (Tick20Val ? 0 : 1);
      digitalWrite(TEST_PIN0, Tick20Val);
    }
    if (++Tick21x >= DAC_VALS) Tick21x = 0;
    dacWrite(DAC2, DACvals[Tick21x]);
#else
    Tick20Val = (Tick20Val ? 0 : 1);
    digitalWrite(TEST_PIN0, Tick20Val);

    if (++Tick21x >= 5) {
      Tick21x = 0;
      Tick21Val = (Tick21Val ? 0 : 1);
      digitalWrite(TEST_PIN1, Tick21Val);
    }
#endif
  }
  xSemaphoreGiveFromISR(Tick2Semaphore, NULL);
}

void setupTick2(size_t sw) {
  if (Tick2Run) {
    Tick2Run = 0;
    delay(1);
  }

  if (Tick2Timer) {
    timerEnd(Tick2Timer);
    Tick2Timer = NULL;
  }

  if (sw) {
    Tick2Semaphore = xSemaphoreCreateBinary();
    Tick2Timer = timerBegin(1,  400, true);
    timerAttachInterrupt(Tick2Timer, &Tick2TimerISR, true);
#ifdef USE_WAVE_DAC
    timerAlarmWrite(Tick2Timer,  10, true);
#else
    timerAlarmWrite(Tick2Timer, 100, true);
#endif
    timerAlarmEnable(Tick2Timer);
    Tick2Run++;
  }
}

// Setup
// --------------------------------------------------------------------
void setup(void) {
  M5.begin();
  M5.Lcd.setBrightness(100);

#if (0)
  pinMode(BUTTON_A_PIN, INPUT_PULLUP);
  pinMode(BUTTON_B_PIN, INPUT_PULLUP);
  pinMode(BUTTON_C_PIN, INPUT_PULLUP);
#endif
  pinMode(ADC_PIN0,  OUTPUT);
  pinMode(ADC_PIN1 , OUTPUT);
  pinMode(TEST_PIN0, OUTPUT);
  pinMode(TEST_PIN1, OUTPUT);
  digitalWrite(TEST_PIN0, 0);
  digitalWrite(TEST_PIN1, 0);

#ifdef USE_WAVE_DAC
  pinMode(DAC2,      OUTPUT);
#endif
  initScreen();
  SetupWaveform(sv_wf_sel);
  setupTick1(hz_unitv[sv_horz_unit]);
}

// Loop
// --------------------------------------------------------------------
void loop() {
  static int buffGetP, buffPutP;

#if (0)
  if ((waveGetRun) && (xSemaphoreTake(Tick1Semaphore, 0) == pdTRUE)) {
    portENTER_CRITICAL(&Tick1TimerMux);
    {
      buffGetP = dmaCountG;
      buffPutP = dmaCountP;
      if ((buffPutP + 1) >= NUMBUFF) buffPutP = 0;
      if (buffPutP == buffGetP) {
        waveGetRun--;
        int point = viewWaves(buffGetP, buffPutP);
        buffGetP += point;
        dmaCountG = ((buffGetP >= NUMBUFF) ? (buffGetP - NUMBUFF) : buffGetP);
        viewWaves(buffGetP, buffPutP);
      }
    }
    portEXIT_CRITICAL(&Tick1TimerMux);
  }
  if (waveGetRun == 0) waveGetRun++;
  M5.update();
  delay(10);
  TouchCheck();
  delay(10);

#else

#if (0)
  if (waveGetRun)
    tft_fillRect(VW_xb + VW_xh - 10, TRIG_INFO_TOP + 2, 10, 10, ixWHITE);

  if ((waveGetRun) && (xSemaphoreTake(Tick1Semaphore, 0) == pdTRUE)) {
    portENTER_CRITICAL(&Tick1TimerMux);
    {
      buffPutP = dmaCountP;
      if (buffPutP >= (NUMBUFF - 1)) {
        waveGetRun--;
        tft_fillRect(VW_xb + VW_xh - 10, TRIG_INFO_TOP + 2, 10, 10, ixGRAY1);
        viewWaves(0, buffPutP);
      }
    }
    portEXIT_CRITICAL(&Tick1TimerMux);
  }
  if (waveGetRun == 0) waveGetRun++;
  M5.update();
  delay(10);
  TouchCheck();
  delay(10);

#else
  if (waveGetRun)
    tft_fillRect(VW_xb + VW_xh - 10, TRIG_INFO_TOP + 2, 10, 10, ixWHITE);

  if ((waveGetRun == 1) && (dmaCountP >= (NUMBUFF - 1))) {
    waveGetRun--;
    do delay(1); while (waveGenStart);
    //digitalWrite(TEST_PIN1, 1);
    tft_fillRect(VW_xb + VW_xh - 10, TRIG_INFO_TOP + 2, 10, 10, ixGRAY1);
    viewWaves(0, dmaCountP);
    //digitalWrite(TEST_PIN1, 0);
    waveGetRun = 1;
    if (hz_unitv[sv_horz_unit] > 10000) delay(15);
  }
  M5.update();
  delay(1);
  TouchCheck();

#endif
#endif
}

void initScreen(void) {
  tft_fillRect(TFT_MIN_X, TFT_MIN_Y, TFT_MAX_X, TFT_MAX_Y, ixBLUE );
#ifdef FUCTION_KEY
  tft_fillRect(TFT_MIN_X, TFT_MAX_Y - 8, TFT_MAX_X, 8, ixBLACK);
#endif
  tft_fillRect(VW_xb    , VW_yb    ,     VW_xh,     VW_yh, ixBLACK);
  tft_drawRect(VW_xb - 1, VW_yb - 1, VW_xh + 3, VW_yh + 3, ixWHITE);
  drawGrid();
  drawGaid();
  drawMenu();

  tft_fillRect(40,  70, 200, 50, ixBLUE);
  tft_drawRect(39,  69, 202, 52, ixWHITE);
  tft_drawText(60,  80, "M5SScope " M5SScope_VERSION, ixWHITE, ixBLUE);
  tft_drawText(60, 102, "(C) 2018 Goji2100.", ixWHITE, ixBLUE);
  delay(1000);

  tft_fillRect(VW_xb    , VW_yb    ,     VW_xh,     VW_yh, ixBLACK);
  tft_drawRect(VW_xb - 1, VW_yb - 1, VW_xh + 3, VW_yh + 3, ixWHITE);
  drawGrid();
}

void drawGrid(void) {
  for (int x =  0; x <= VW_xh; x += 20)
    tft_drawVDotLine(VW_xb + x, VW_yb, VW_yh, (x == (VW_xh / 2)) ? ixYELLOW : ixGRAY2);
  for (int y =  0; y <= VW_yh; y += 20)
    tft_drawHDotLine(VW_xb, VW_yb + y, VW_xh, (y == (VW_yh / 2)) ? ixYELLOW : ixGRAY2);
  tft_drawVDotLine(VW_xb + (VW_xh / 2), VW_yb, VW_yh, ixYELLOW);
}

//  View Wave
//---------------------------------------------------------------------
int viewWaves(int GetP, int PutP) {
  uint16_t i, s = 0;
  uint16_t tr1, tr2, tv1, tv2;

  if (sv_tr_sel != SV_TR_SEL_NONE) {
    tv1 = tv2 = 0;
    tr1 = (sv_tr_levn[0] * 4096) / 5000;
    tr2 = (sv_tr_levn[1] * 4096) / 5000;
    s = (VW_xh / 2) + (sv_hz_posn / (hz_unitn[sv_horz_unit] / 20));

    for (; s < (VW_xh + (VW_xh / 2)); s++) {
      tv1 = tv2 = 0;
      if ((sv_tr_sel & SV_TR_SEL_CH1) ||
          (sv_tr_sel & SV_TR_SEL_AND)) {
        switch (sv_tr_type[0]) {
          case SV_TR_TYPE_UP:
            if ((wavebuf[0][s]     < tr1) &&
                (wavebuf[0][s + 1] > tr1))
              tv1++;
            break;
          case SV_TR_TYPE_DWN:
            if ((wavebuf[0][s]     > tr1) &&
                (wavebuf[0][s + 1] < tr1))
              tv1++;
            break;
          case SV_TR_TYPE_HI:
            if (wavebuf[0][s]     > tr1)
              tv1++;
            break;
          case SV_TR_TYPE_LOW:
            if (wavebuf[0][s]     < tr1)
              tv1++;
            break;
        }
      }
      if ((sv_tr_sel & SV_TR_SEL_CH2) ||
          ((sv_tr_sel & SV_TR_SEL_AND) && tv1)) {
        switch (sv_tr_type[1]) {
          case SV_TR_TYPE_UP:
            if ((wavebuf[1][s]     < tr2) &&
                (wavebuf[1][s + 1] > tr2))
              tv2++;
            break;
          case SV_TR_TYPE_DWN:
            if ((wavebuf[1][s]     > tr2) &&
                (wavebuf[1][s + 1] < tr2))
              tv2++;
            break;
          case SV_TR_TYPE_HI:
            if (wavebuf[1][s]     > tr2)
              tv2++;
            break;
          case SV_TR_TYPE_LOW:
            if (wavebuf[1][s]     < tr2)
              tv2++;
            break;
        }
      }
      if (sv_tr_sel & SV_TR_SEL_AND) {
        if (tv1 && tv2)
          break;
      } else {
        if (((sv_tr_sel & SV_TR_SEL_CH1) && tv1) ||
            ((sv_tr_sel & SV_TR_SEL_CH2) && tv2))
          break;
      }
    }
    if (s < (VW_xh + (VW_xh / 2)))
      s -= ((VW_xh / 2) + (sv_hz_posn / (hz_unitn[sv_horz_unit] / 20)));
  }

  if (s < VW_xh) {
    for (i = 0; i <= VW_xh; i++) {
      wavebuf[0][i] = (wavebuf[0][s + i] * 2000) / 4096 / (vd_seln[sv_ch_vdiv[0]] / 50);
      wavebuf[0][i] = ((wavebuf[0][i] + (VW_yh - sv_ch_poss[0])) > (VW_yh - 1)) ? (VW_yh - 1) : (wavebuf[0][i] + (VW_yh - sv_ch_poss[0]));
      wavebuf[1][i] = (wavebuf[1][s + i] * 2000) / 4096 / (vd_seln[sv_ch_vdiv[1]] / 50);
      wavebuf[1][i] = ((wavebuf[1][i] + (VW_yh - sv_ch_poss[1])) > (VW_yh - 1)) ? (VW_yh - 1) : (wavebuf[1][i] + (VW_yh - sv_ch_poss[1]));
    }
    for (i = 0; i <= VW_xh; i++) {
      if (sv_ch_sel & SV_CH1) {
        wavebuf[0][i] = (VW_yh - 1) - wavebuf[0][i];
        waverec[0][0][i] = ((wavebuf[0][i] >= VW_yh) ? (VW_yh - 1) : wavebuf[0][i]) + VW_yb;
      }
      if (sv_ch_sel & SV_CH2) {
        wavebuf[1][i] = (VW_yh - 1) - wavebuf[1][i];
        waverec[1][0][i] = ((wavebuf[1][i] >= VW_yh) ? (VW_yh - 1) : wavebuf[1][i]) + VW_yb;
      }
    }
    drawWave(ch1_color, ch2_color);
  }
  return (s + VW_xh);
}

#define VALMIN(a, b) (((a)<(b))?(a):(b))
#define VALMAX(a, b) (((a)>(b))?(a):(b))

//  Draw Wave
//---------------------------------------------------------------------
void drawWave(uint16_t ch1color, uint16_t ch2color) {
  static uint16_t wave_r1rep, wave_r2rep;
  int i;

  for (i = 0; i < VW_xh; i++) {
    if ((wave_r1rep) && (wave_r2rep)) {
      tft_drawVDotLineRep(VW_xb + i,
                          VALMIN(VALMIN(waverec[0][1][i], waverec[1][1][i]), VALMIN(waverec[0][1][i + 1], waverec[1][1][i + 1])),
                          VALMAX(VALMAX(waverec[0][1][i], waverec[1][1][i]), VALMAX(waverec[0][1][i + 1], waverec[1][1][i + 1])),
                          ixBLACK);
    } else {
      if (wave_r2rep)
        tft_drawVDotLineRep(VW_xb + i, waverec[1][1][i], waverec[1][1][i + 1], ixBLACK);
      if (wave_r1rep)
        tft_drawVDotLineRep(VW_xb + i, waverec[0][1][i], waverec[0][1][i + 1], ixBLACK);
    }
    if (sv_ch_sel & SV_CH2)
      tft_drawVDotLineRep(VW_xb + i, waverec[1][1][i] = waverec[1][0][i], waverec[1][0][i + 1], ch2color);
    if (sv_ch_sel & SV_CH1)
      tft_drawVDotLineRep(VW_xb + i, waverec[0][1][i] = waverec[0][0][i], waverec[0][0][i + 1], ch1color);
  }
  waverec[0][1][i] = waverec[0][0][i];
  waverec[1][1][i] = waverec[1][0][i];
  wave_r1rep = (sv_ch_sel & SV_CH1); // ? hz_unitv[sv_horz_unit] : 0;
  wave_r2rep = (sv_ch_sel & SV_CH2); // ? hz_unitv[sv_horz_unit] : 0;
}



//  Menu
//---------------------------------------------------------------------

typedef struct {
  uint16_t px, py;          // Menu Button Pos(Top, Left)
  uint16_t wx, wy;          // Button Size
  uint16_t color;           // Button Color
  char  *ms;                // Menu Strings
  void  (*fn)(uint16_t n);  // On Click()
} MENU_ST;

void MenuSelMark(MENU_ST *menu_s, uint16_t scolor, int line);
void drawMenuE(MENU_ST *menu_s);
void drawMenuS(MENU_ST *menu_s);
MENU_ST *MenuCheck(MENU_ST *menu_s, uint16_t x, uint16_t y);

#define MENU_MENU_COLOR ixWHITE
#define MENU_SELC_COLOR ixYELLOW
#define MENU_EXIT_COLOR ixGRAY1

#define MENU_SEL_LR     1
#define MENU_SEL_UD     2

char ch1__smenu[] = "\004\007\007\010   1V\0\004\025\007\010 0.0V\0";
char ch2__smenu[] = "\004\007\007\010   1V\0\004\025\007\010 0.0V\0";
char horz_smenu[] = "\004\007\007\010  1ms\0\004\025\007\010  0ms\0";
char trig_smenu[] = "\010\016\007\010Menu\0";
char util_smenu[] = "\010\016\007\010Menu\0";
char dumy_smenu[] = "\000\000\000\000\0";
char *trig_slopes[] = { " Free ", "\x13\x2F\x10", "\x1F\x5C\x5F", "\x13\x11\x5F", "\x1F\x12\x10" };

char pressBtnA = 0;

void TouchCheck(void) {
  if (pressBtnA) {
    if ((sv_dp_sel != SV_DP_SEL_NONE) && M5.BtnA.pressedFor(800)) {
      dumpScreen(sv_dp_sel);
      pressBtnA = 0;
    } else {
      if (M5.BtnA.wasReleased()) {
        MenuSelect(SW_SELECT);
        pressBtnA = 0;
      }
    }
  }
  if (M5.BtnA.wasPressed()) pressBtnA = 1;
  if (M5.BtnB.wasPressed()) MenuSelect(SW_LEFT);
  if (M5.BtnC.wasPressed()) MenuSelect(SW_RIGHT);
}

void drawGaid(void) {
  uint16_t i, px;
  char  estr[16];

  //	Trigger Info
  //-----------------------------------------------------------------
  tft_fillRect(TFT_MIN_X, TFT_MIN_Y, TFT_MAX_X, 2 + 12, ixBLUE);
  px = 18 + (tft_drawText( 18, TRIG_INFO_TOP, (char *)tr_sels[sv_tr_sel], ch1_color, ixGRAY1) * 8) + 4;
  tft_drawText(VW_xb + 254, TRIG_INFO_TOP, M5SScope_VERSION, ixWHITE, ixBLUE);

  if ((sv_tr_sel & SV_TR_SEL_CH1) || (sv_tr_sel & SV_TR_SEL_AND)) {
    px += (tft_drawText( px, TRIG_INFO_TOP, (char *)trig_slopes[sv_tr_type[0]], ixGREEN, ixGRAY1) * 8) + 4;
    l2Vstr(estr, sv_tr_levn[0]);
    px += (tft_drawText( px, TRIG_INFO_TOP, (char *)estr, ((estr[0] == '-') ? ixRED : ixGREEN), ixGRAY1) * 8) + 8;
  }

  if ((sv_tr_sel & SV_TR_SEL_CH2) || (sv_tr_sel & SV_TR_SEL_AND)) {
    px += (tft_drawText( px, TRIG_INFO_TOP, (char *)trig_slopes[sv_tr_type[1]], ixGREEN, ixGRAY1) * 8) + 4;
    l2Vstr(estr, sv_tr_levn[1]);
    px += (tft_drawText( px, TRIG_INFO_TOP, (char *)estr, ((estr[0] == '-') ? ixRED : ixGREEN), ixGRAY1) * 8) + 8;
  }

#if (TRIG_INFO_TOP == 0)
  tft_drawFastHLine(VW_xb - 1, VW_yb - 1, VW_xh + 2, ixWHITE);
#endif

  for (i = 0; i < 5; i++) {
    ch1__smenu[ 4 + i] = vd_sels[sv_ch_vdiv[0]][i];	// CH1 div
    ch2__smenu[ 4 + i] = vd_sels[sv_ch_vdiv[1]][i];	// CH2 div
    horz_smenu[ 4 + i] = hz_units[sv_horz_unit][i];	// Horz div
  }

  ch1__smenu[12] = (l2Vstr5(ch1__smenu + 14, sv_ch_posn[0]) == -1) ? ixRED : ixWHITE;
  ch2__smenu[12] = (l2Vstr5(ch2__smenu + 14, sv_ch_posn[1]) == -1) ? ixRED : ixWHITE;
  horz_smenu[12] = (l2Sstr5(horz_smenu + 14, sv_hz_posn   ) == -1) ? ixRED : ixWHITE;

  //	Vertical Marker
  //-----------------------------------------------------------------
  calcVpos();
  tft_fillRect(TFT_MIN_X, TFT_MIN_Y, VW_xb - 1, VW_yb + VW_yh + VW_yb, ixBLUE);

  if ((sv_tr_sel & SV_TR_SEL_CH2) || (sv_tr_sel & SV_TR_SEL_AND))
    tft_drawICON(  1, VW_yb + sv_tr_levs[1] - 5, 13, 11, (uint16_t *)ICONs[0], ch2_color, ixBLUE);	// 'T>'
  if ((sv_tr_sel & SV_TR_SEL_CH1) || (sv_tr_sel & SV_TR_SEL_AND))
    tft_drawICON(  1, VW_yb + sv_tr_levs[0] - 5, 13, 11, (uint16_t *)ICONs[0], ch1_color, ixBLUE);	// 'T>'
  if (sv_ch_sel & SV_CH2)
    tft_drawICON(  1, VW_yb + sv_ch_poss[1] - 5, 13, 11, (uint16_t *)ICONs[2], ch2_color, ixBLUE);	// '2>'
  if (sv_ch_sel & SV_CH1)
    tft_drawICON(  1, VW_yb + sv_ch_poss[0] - 5, 13, 11, (uint16_t *)ICONs[1], ch1_color, ixBLUE);	// '1>'
}

void MenuSelMark(MENU_ST *menu_s, uint16_t scolor, int line) {
  if (scolor == MENU_MENU_COLOR) {
    tft_drawRect( menu_s->px     , (menu_s->py) + 2,  menu_s->wx     , (menu_s->wy) - 4, MENU_EXIT_COLOR);
    tft_drawRect((menu_s->px) + 2,  menu_s->py     , (menu_s->wx) - 4,  menu_s->wy     , MENU_EXIT_COLOR);
    tft_drawRect( menu_s->px, menu_s->py, menu_s->wx, menu_s->wy, ixWHITE);
  } else {
    tft_drawRect( menu_s->px, menu_s->py, menu_s->wx, menu_s->wy, scolor);
    if (line == MENU_SEL_LR) {
      tft_drawRect((menu_s->px) + 2,  menu_s->py     , (menu_s->wx) - 4,  menu_s->wy     , scolor);
#ifdef FUCTION_KEY
      tft_drawICON( 64, 232, 16, 8, (uint16_t *)ICONs[3], ixWHITE, ixBLACK);
      tft_drawICON(158, 232, 16, 8, (uint16_t *)ICONs[4], ixWHITE, ixBLACK);
      tft_drawICON(251, 232, 16, 8, (uint16_t *)ICONs[5], ixWHITE, ixBLACK);
#endif
    } else {
      tft_drawRect( menu_s->px     , (menu_s->py) + 2,  menu_s->wx     , (menu_s->wy) - 4, scolor);
#ifdef FUCTION_KEY
      tft_drawICON( 64, 232, 16, 8, (uint16_t *)ICONs[6], ixWHITE, ixBLACK);
      tft_drawICON(158, 232, 16, 8, (uint16_t *)ICONs[7], ixWHITE, ixBLACK);
      tft_drawICON(251, 232, 16, 8, (uint16_t *)ICONs[8], ixWHITE, ixBLACK);
#endif
    }
  }
}

//	Main menu
//---------------------------------------------------------------------
MENU_ST menu_main[] = {
  { VW_xb + (49 * 0), VW_yb + 177, 46, 40, ixGRAY1, ch1__smenu, MenuCH12 },
  { VW_xb + (49 * 1), VW_yb + 177, 46, 40, ixGRAY1, ch2__smenu, MenuCH12 },
  { VW_xb + (49 * 2), VW_yb + 177, 46, 40, ixGRAY1, horz_smenu, MenuHorz },
  { VW_xb + (49 * 3), VW_yb + 177, 46, 40, ixGRAY1, trig_smenu, MenuTrig },
  { VW_xb + (49 * 4), VW_yb + 177, 46, 40, ixGRAY1, util_smenu, MenuUTIL },
  { 0, 0, 0, 0, 0, 0 }
};

//	Sub menu (Empty)
//---------------------------------------------------------------------
MENU_ST menu_none[] = {
  { VW_xb + 248, VW_yb + (44 * 0) - 1, 52, 41, ixGRAY1, dumy_smenu, 0  },
  { VW_xb + 248, VW_yb + (44 * 1)    , 52, 40, ixGRAY1, dumy_smenu, 0  },
  { VW_xb + 248, VW_yb + (44 * 2)    , 52, 40, ixGRAY1, dumy_smenu, 0  },
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, dumy_smenu, 0  },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, dumy_smenu, 0  },
  { 0, 0, 0, 0, 0, 0 }
};

//	Draw Menu
//---------------------------------------------------------------------
MENU_ST *menu_top = (MENU_ST*)menu_main;
MENU_ST *menu_sub = (MENU_ST*)menu_none;

int menu_top_pos =  0;
int menu_sub_pos = -1;
int menu_sub_sel[5] = { 0, 0, 0, 0, 0 };

void drawMenu(void) {
  //	Main Menu
  //-----------------------------------------------------------------
  for (int i = 0; i < 5; i++) {
    tft_fillRect(VW_xb + (i * 49), VW_yb + 164, 46, 12, ixGRAY2);
    tft_drawFastHLine(VW_xb + (i * 49), VW_yb + 164 + 12, 46, ixWHITE);
  }
  tft_drawText(VW_xb + ( 0 * 49) + 12, VW_yb + 164, "CH1"   , ch1_color, ixBLACK1);
  tft_drawText(VW_xb + ( 1 * 49) + 12, VW_yb + 164, "CH2"   , ch2_color, ixBLACK1);
  tft_drawText(VW_xb + ( 2 * 49) +  8, VW_yb + 164, "Horz"  , ixWHITE,   ixBLACK1);
  tft_drawText(VW_xb + ( 3 * 49) +  8, VW_yb + 164, "Trig"  , ixWHITE,   ixBLACK1);
  tft_drawText(VW_xb + ( 4 * 49) +  4, VW_yb + 164, "Util"  , ixWHITE,   ixBLACK1);
  drawMenuE(menu_top = (MENU_ST*)menu_main);
  MenuSelMark(menu_top + menu_top_pos, ((menu_sub_pos != -1) ? MENU_MENU_COLOR : MENU_SELC_COLOR), MENU_SEL_LR);
  MenuCH12(0); // 0:CH1
}

void drawMenuE(MENU_ST *menu_s) {
  for (; menu_s->color != 0; menu_s++)
    drawMenuS(menu_s);
}

void drawMenuS(MENU_ST *menu_s) {
  uint16_t pox, poy, x, y, c, b, l;

  tft_fillRect(pox = menu_s->px, poy = menu_s->py, menu_s->wx, menu_s->wy, menu_s->color);
  for (char *s = menu_s->ms; *s; s++) {
    x = pox + *s++;
    y = poy + *s++;
    c = *s++;
    b = *s++;
    if (*s == '\0') break;
    l = tft_drawText(x, y, s, c, b);
    s += l;
  }
}

//	Button Check
//---------------------------------------------------------------------
void MenuSelect(int btn) {

  // Main menu select
  // ----------------
  if (menu_sub_pos == -1) {
    MENU_ST *menu_s = (menu_top + menu_top_pos);
    if (btn == SW_SELECT) {
      MenuSelMark(menu_s, MENU_MENU_COLOR, MENU_SEL_LR); // Enter sub menu
      if (menu_s->fn) (menu_s->fn)(menu_top_pos); // Display sub menu
      menu_sub_pos = menu_sub_sel[menu_top_pos];  // Restore sub menu position
      MenuSelMark(menu_sub + menu_sub_pos, MENU_SELC_COLOR, MENU_SEL_UD); // Sub menu select
      return;
    }

    // Main menu Left / Right
    // ----------------------
    MenuSelMark(menu_s, MENU_EXIT_COLOR, MENU_SEL_LR);  // Deselect
    if (btn == SW_LEFT) menu_top_pos--; else menu_top_pos++;
    if (menu_top_pos < (    0)) menu_top_pos = (5 - 1);
    if (menu_top_pos > (5 - 1)) menu_top_pos = 0;
    menu_s = menu_top + menu_top_pos;
    MenuSelMark(menu_s, MENU_SELC_COLOR, MENU_SEL_LR);  // Select
    if (menu_s->fn) (menu_s->fn)(menu_top_pos);
    return;
  }

  // Sub Menu select
  // ---------------
  MENU_ST* menu_s = (menu_sub + menu_sub_pos);
  if (btn == SW_SELECT) {
    menu_sub_sel[menu_top_pos] = menu_sub_pos;
    if (menu_s->fn) (menu_s->fn)(menu_sub_pos);
    MenuSelMark(menu_s, MENU_SELC_COLOR, MENU_SEL_UD);
    MenuSelMark(menu_top + menu_top_pos, MENU_MENU_COLOR, MENU_SEL_UD);
    return;
  }

  // Sub menu Up / Down
  // ------------------
  if (((menu_sub_pos > (    0)) && (btn == SW_LEFT)) ||
      ((menu_sub_pos < (5 - 1)) && (btn == SW_RIGHT))) {
    MenuSelMark(menu_s, MENU_EXIT_COLOR, MENU_SEL_UD);
    if (btn == SW_LEFT) menu_sub_pos--; else menu_sub_pos++;
    MenuSelMark(menu_sub + menu_sub_pos, MENU_SELC_COLOR, MENU_SEL_UD);
    return;
  }

  // Exit Sub Menu
  // -------------
  if (((menu_sub_pos <= (    0)) && (btn == SW_LEFT)) ||
      ((menu_sub_pos >= (5 - 1)) && (btn == SW_RIGHT))) {
    MenuSelMark(menu_s, MENU_EXIT_COLOR, MENU_SEL_UD);
    MenuSelMark(menu_top + menu_top_pos, MENU_SELC_COLOR, MENU_SEL_LR);
    menu_sub_pos = -1;
    return;
  }
  return;
}

MENU_ST *MenuCheck(MENU_ST *menu_s, uint16_t x, uint16_t y) {
  if ((((menu_s->px) <= x) && (x <= (menu_s->px + menu_s->wx))) &&
      (((menu_s->py) <= y) && (y <= (menu_s->py + menu_s->wy))))
    return menu_s;
  return 0;
}


//	Sub menu (CH1 / CH2)
//---------------------------------------------------------------------
char chns_chOn [] = "\016\010\007\010CH1\0\012\026\010\007 ON \0";
char chns_chOff[] = "\016\010\007\010CH1\0\006\026\010\007 OFF \0";
char chns_volDN[] = "\016\010\007\010Div\0\024\026\007\010UP\0";
char chns_volUP[] = "\016\010\007\010Div\0\012\026\007\010DOWN\0";
char chns_posUP[] = "\016\010\007\010Pos\0\024\026\007\010UP\0";
char chns_posDN[] = "\016\010\007\010Pos\0\012\026\007\010DOWN\0";

MENU_ST menu_chn[] = {
  { VW_xb + 248, VW_yb + (44 * 0) - 1, 52, 41, ixGRAY1, chns_chOn , SMenuCHon   },
  { VW_xb + 248, VW_yb + (44 * 1)    , 52, 40, ixGRAY1, chns_volDN, SMenuCHvdiv },
  { VW_xb + 248, VW_yb + (44 * 2)    , 52, 40, ixGRAY1, chns_volUP, SMenuCHvdiv },
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, chns_posUP, SMenuCHvpos },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, chns_posDN, SMenuCHvpos },
  { 0, 0, 0, 0, 0, 0 }
};


//	CH1 / CH2 Menu
//---------------------------------------------------------------------
uint16_t sv_menuCHn = 0;		// Menu selected Channel (CH1:0 / CH2:1)

void MenuCH12(uint16_t ch) {
  if (ch == 0) {				// CH1 = 0, CH2 = 1
    sv_menuCHn = 0;
    menu_chn[0].ms = (sv_ch_sel & SV_CH1) ? chns_chOn : chns_chOff;
  } else {
    sv_menuCHn = 1;
    menu_chn[0].ms = (sv_ch_sel & SV_CH2) ? chns_chOn : chns_chOff;
  }
  chns_chOn [6] = chns_chOff[6] = (ch + '1');
  drawMenuE(menu_sub = menu_chn);
}

//	Display ON / OFF
//---------------------------------------------------------
void SMenuCHon(uint16_t none) {
  if (sv_menuCHn == 0) sv_ch_sel ^= SV_CH1;
  if (sv_menuCHn == 1) sv_ch_sel ^= SV_CH2;
  drawGaid();
  MenuCH12(sv_menuCHn);
}

//	Vertical Div UP / DOWN
//---------------------------------------------------------
void SMenuCHvdiv(uint16_t updown) {
  if (updown == 2) {			// Down = 1, Up = 2
    if (sv_ch_vdiv[sv_menuCHn] < (VT_UNIT_MAX - 1)) sv_ch_vdiv[sv_menuCHn]++;
  } else {
    if (sv_ch_vdiv[sv_menuCHn] > (              0)) sv_ch_vdiv[sv_menuCHn]--;
  }
  sv_ch_posn[sv_menuCHn] /= (vd_seln[sv_ch_vdiv[sv_menuCHn]] / 20);
  sv_ch_posn[sv_menuCHn] *= (vd_seln[sv_ch_vdiv[sv_menuCHn]] / 20);
  drawGaid();
  drawMenuS((sv_menuCHn == 0) ? (MENU_ST *)&menu_main[0] : (MENU_ST *)&menu_main[1]);
}

//	Vertical Pos UP / DOWN
//---------------------------------------------------------
void SMenuCHvpos(uint16_t updown) {
  if (updown == 3) {			// UP = 3, DOWN = 4
    sv_ch_posn[sv_menuCHn] += (vd_seln[sv_ch_vdiv[sv_menuCHn]] / 20);
    if (sv_ch_posn[sv_menuCHn] > (vd_seln[sv_ch_vdiv[sv_menuCHn]] * 8))
      sv_ch_posn[sv_menuCHn] = (vd_seln[sv_ch_vdiv[sv_menuCHn]] * 8);
  } else {
    sv_ch_posn[sv_menuCHn] -= (vd_seln[sv_ch_vdiv[sv_menuCHn]] / 20);
    if (sv_ch_posn[sv_menuCHn] < 0L)
      if (sv_ch_posn[sv_menuCHn] < (0L - (vd_seln[sv_ch_vdiv[sv_menuCHn]] * 8)))
        sv_ch_posn[sv_menuCHn] = (0L - (vd_seln[sv_ch_vdiv[sv_menuCHn]] * 8));
  }
  drawGaid();
  drawMenuS((sv_menuCHn == 0) ? (MENU_ST *)&menu_main[0] : (MENU_ST *)&menu_main[1]);
}

uint16_t calcVposs(long n, long f) {
  n = (n + (f * 4)) / (f / 20);
  if (n > (VW_yh - 1)) n = (VW_yh - 1); else if (n < 0) n = 0;
  return (VW_yh - n);
}

void calcVpos(void) {
  for (int i = 0; i <= 1; i++) {
    sv_ch_poss[i] = calcVposs(sv_ch_posn[i], vd_seln[sv_ch_vdiv[i]]);
    sv_tr_levs[i] = calcVposs(sv_tr_levn[i] + sv_ch_posn[i], vd_seln[sv_ch_vdiv[i]]);
  }
}

//	Sub menu (Horizontal)
//---------------------------------------------------------------------
const char horz_divUP[] = "\016\010\007\010Div\0\024\026\007\010UP\0";
const char horz_divDN[] = "\016\010\007\010Div\0\012\026\007\010DOWN\0";
const char horz_posLF[] = "\016\010\007\010Pos\0\012\026\007\010 <<\0";
const char horz_posRT[] = "\016\010\007\010Pos\0\012\026\007\010 >>\0";

const MENU_ST menu_horz[] = {
  { VW_xb + 248, VW_yb + (44 * 0) - 1, 52, 41, ixGRAY1, (char*)horz_divUP, SMenuHZsdiv },
  { VW_xb + 248, VW_yb + (44 * 1)    , 52, 40, ixGRAY1, (char*)horz_divDN, SMenuHZsdiv },
  { VW_xb + 248, VW_yb + (44 * 2)    , 52, 40, ixGRAY1, (char*)horz_posLF, SMenuHZhpos },
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, (char*)horz_posRT, SMenuHZhpos },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, (char*)dumy_smenu, 0           },
  { 0, 0, 0, 0, 0, 0 }
};

//	Horizontal Menu
//---------------------------------------------------------------------
void MenuHorz(uint16_t none) {
  drawMenuE(menu_sub = (MENU_ST *)menu_horz);
}

void SMenuHZsdiv(uint16_t updown) {
  if (updown == 0) {		// UP = 0, DOWN = 1
    if (sv_horz_unit > 0)
      setupTick1(hz_unitv[--sv_horz_unit]);
  } else {
    if (sv_horz_unit < (HZ_UNIT_MAX - 1))
      setupTick1(hz_unitv[++sv_horz_unit]);
  }
  SMenuHZhposchk();
  drawGaid();
  drawMenuS((MENU_ST *)&menu_main[2]);
}

void SMenuHZhpos(uint16_t lr) {
  if (lr == 2) {			// Left = 2, Right = 3
    sv_hz_posn -= (hz_unitn[sv_horz_unit] / 20);
  } else {
    sv_hz_posn += (hz_unitn[sv_horz_unit] / 20);
  }
  SMenuHZhposchk();
  drawGaid();
  drawMenuS((MENU_ST *)&menu_main[2]);
}

void SMenuHZhposchk(void) {
  sv_hz_posn /= (hz_unitn[sv_horz_unit] / 20);
  sv_hz_posn *= (hz_unitn[sv_horz_unit] / 20);

  if (sv_hz_posn > (hz_unitn[sv_horz_unit] * 6)) {
    sv_hz_posn = (hz_unitn[sv_horz_unit] * 6);
  } else {
    if (sv_hz_posn < 0L)
      if (sv_hz_posn < (0L - (hz_unitn[sv_horz_unit] * 6)))
        sv_hz_posn = (0L - (hz_unitn[sv_horz_unit] * 6));
  }
}

//	Sub menu (Trigger)
//---------------------------------------------------------------------
char *trig_selch[] = {
  "\012\010\007\010Mode\0\002\026\010\007 Free \0",
  "\012\010\007\010Mode\0\006\026\010\007 CH1 \0",
  "\012\010\007\010Mode\0\006\026\010\007 CH2 \0",
  "\012\010\007\010Mode\0\006\026\010\007 1|2 \0",
  "\012\010\007\010Mode\0\006\026\010\007 1&2 \0"
};

char *trig_slope[] = {
  "\006\010\007\010Slope\0\002\026\010\007 Free \0",			  // Free
  "\006\010\007\010Slope\0\016\026\010\007\x13\x2F\x10\0",	// Rising
  "\006\010\007\010Slope\0\016\026\010\007\x1F\x5C\x5F\0",	// Falling
  "\006\010\007\010Slope\0\016\026\010\007\x13\x11\x5F\0",	// Above
  "\006\010\007\010Slope\0\016\026\010\007\x1F\x12\x10\0"   // Below
};

char trig_setup[] = "\006\010\007\010Setup\0\006\026\010\007 CH1 \0";
char trig_divUP[] = "\010\010\007\010Level\0\024\026\007\010UP\0";
char trig_divDN[] = "\010\010\007\010Level\0\012\026\007\010DOWN\0";

MENU_ST menu_trig[] = {
  { VW_xb + 248, VW_yb + (44 * 0) - 1, 52, 41, ixGRAY1, dumy_smenu, SMenuTrg12  },
  { VW_xb + 248, VW_yb + (44 * 1)    , 52, 40, ixGRAY1, trig_setup, SMenuTrgSet },
  { VW_xb + 248, VW_yb + (44 * 2)    , 52, 40, ixGRAY1, dumy_smenu, SMenuTrgSLP },
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, trig_divUP, SMenuTrgLVL },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, trig_divDN, SMenuTrgLVL },
  { 0, 0, 0, 0, 0, 0 }
};


//	Trigger Menu
//---------------------------------------------------------------------
uint16_t sv_menuTRn = 0;  // Menu selected Channel (CH1:0 / CH2:1)

void MenuTrig(uint16_t none) {
  menu_trig[0].ms = trig_selch[sv_tr_sel];
  menu_trig[2].ms = trig_slope[sv_tr_type[sv_menuTRn]];
  trig_setup[17] = (sv_menuTRn == 0) ? '1' : '2';
  drawMenuE(menu_sub = menu_trig);
  drawGaid();
}

void SMenuTrg12(uint16_t none) {
  if (++sv_tr_sel > SV_TR_SEL_AND) sv_tr_sel = SV_TR_SEL_NONE;
  MenuTrig(0);
}

void SMenuTrgSet(uint16_t none) {
  sv_menuTRn = (sv_menuTRn == 0) ? 1 : 0;
  MenuTrig(0);
}

void SMenuTrgSLP(uint16_t none) {
  if (++sv_tr_type[sv_menuTRn] > SV_TR_TYPE_LOW) sv_tr_type[sv_menuTRn] = SV_TR_TYPE_UP;
  MenuTrig(0);
}

void SMenuTrgLVL(uint16_t updown) {
  if (updown == 3) {		// UP = 3, DOWN = 4
    sv_tr_levn[sv_menuTRn] += (vd_seln[sv_ch_vdiv[sv_menuTRn]] / 20);
  } else {
    sv_tr_levn[sv_menuTRn] -= (vd_seln[sv_ch_vdiv[sv_menuTRn]] / 20);
  }
  if (sv_tr_levn[sv_menuTRn] > (vd_seln[sv_ch_vdiv[sv_menuTRn]] * 8)) {
    sv_tr_levn[sv_menuTRn] = (vd_seln[sv_ch_vdiv[sv_menuTRn]] * 8);
  } else {
    if (sv_tr_levn[sv_menuTRn] < 0L)
      if (sv_tr_levn[sv_menuTRn] < (0L - (vd_seln[sv_ch_vdiv[sv_menuTRn]] * 8)))
        sv_tr_levn[sv_menuTRn] = (0L - (vd_seln[sv_ch_vdiv[sv_menuTRn]] * 8));
  }
  drawGaid();
}

void SMenuWavef(uint16_t none);
void SMenuDraww(uint16_t none);

//  Sub menu (Utility)
//---------------------------------------------------------------------
char *util_wavef[] = {
  "\012\010\007\010Wave\0\002\026\010\007 None \0",         // None
#ifdef USE_WAVE_DAC
  "\012\010\007\010Wave\0\016\026\010\007\x8A\x8B\x8C\0",   // rect
  "\012\010\007\010Wave\0\016\026\010\007\x87\x88\x89\0",   // step
  "\012\010\007\010Wave\0\016\026\010\007\x84\x85\x86\0",   // saw
  "\012\010\007\010Wave\0\016\026\010\007\x81\x82\x83\0"    // sin
#else
  "\012\010\007\010Wave\0\016\026\010\007\x8A\x8B\x8C\0"    // rect
#endif
};

#ifdef DUMP_SCREEN
char *util_dumpp[] = {
  "\012\010\007\010Dump\0\002\026\010\007 None \0",         // None
  "\012\010\007\010Dump\0\006\026\010\007 BMP \0",          // BMP
  "\012\010\007\010Dump\0\002\026\010\007 UART \0"          // UART
};
#endif

MENU_ST menu_util[] = {
  { VW_xb + 248, VW_yb + (44 * 0) - 1, 52, 41, ixGRAY1, dumy_smenu, SMenuWavef  },
  { VW_xb + 248, VW_yb + (44 * 1)    , 52, 40, ixGRAY1, dumy_smenu, 0           },
  { VW_xb + 248, VW_yb + (44 * 2)    , 52, 40, ixGRAY1, dumy_smenu, 0           },
#ifdef DUMP_SCREEN
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, dumy_smenu, 0           },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, dumy_smenu, SMenuDumpp  },
#else
  { VW_xb + 248, VW_yb + (44 * 3)    , 52, 40, ixGRAY1, dumy_smenu, 0           },
  { VW_xb + 248, VW_yb + (44 * 4)    , 52, 41, ixGRAY1, dumy_smenu, 0           },
#endif
  { 0, 0, 0, 0, 0, 0 }
};

//  Test Wave Menu
//---------------------------------------------------------------------
void MenuUTIL(uint16_t none) {
  menu_util[0].ms = util_wavef[sv_wf_sel];
#ifdef DUMP_SCREEN
  menu_util[4].ms = util_dumpp[sv_dp_sel];
#endif
  drawMenuE(menu_sub = menu_util);
  drawGaid();
}

void SMenuWavef(uint16_t none) {
#ifdef USE_WAVE_DAC
  if (++sv_wf_sel > SV_WF_SEL_SIN ) sv_wf_sel = SV_WF_SEL_NONE;
#else
  if (++sv_wf_sel > SV_WF_SEL_RECT) sv_wf_sel = SV_WF_SEL_NONE;
#endif
  SetupWaveform(sv_wf_sel);
  MenuUTIL(0);
}

#ifdef DUMP_SCREEN
void SMenuDumpp(uint16_t none) {
  if (++sv_dp_sel > SV_DP_SEL_UART) sv_dp_sel = SV_DP_SEL_NONE;
  MenuUTIL(0);
}
#endif

void SetupWaveform(uint16_t fn) {
  setupTick2(0);
  if (fn != SV_WF_SEL_NONE) {
#ifdef USE_WAVE_DAC
    for (int i = 0; i < DAC_VALS; i++) {

#define DAC_VOL ((3300 * 255) / 3300)
#if (1)
      if      (fn == SV_WF_SEL_RECT) DACvals[i] = ((i < (DAC_VALS / 2)) ? 0 : DAC_VOL);
      else if (fn == SV_WF_SEL_SAW)  DACvals[i] = ((i * DAC_VOL) / DAC_VALS);
      else if (fn == SV_WF_SEL_STEP) DACvals[i] = (((i / 20) * 25 * DAC_VOL) / DAC_VALS);
      else if (fn == SV_WF_SEL_SIN)  DACvals[i] = (uint8_t)((DAC_VOL / 2) + ((DAC_VOL / 2.0) * sinf(6.283032 * i / DAC_VALS)) + 0.8);
#else
      if      (fn == SV_WF_SEL_RECT) DACvals[i] = ((i < (DAC_VALS / 2)) ? 0 : 232);
      else if (fn == SV_WF_SEL_SAW)  DACvals[i] = ((i * 233) / DAC_VALS);
      else if (fn == SV_WF_SEL_STEP) DACvals[i] = (((i / 20) * 25 * 232) / DAC_VALS);
      else if (fn == SV_WF_SEL_SIN)  DACvals[i] = (uint8_t)(114 + (115.0 * sinf(6.283032 * i / DAC_VALS)) + 0.8);
#endif

    }
    setupTick2(1);
  }
#endif
}

void i2Str(char *s, uint16_t n) {
  *s = ' ', *(s + 6) = '\0';  // "-12345\0"
  for (int i = 5; i > 0; i--) {
    *(s + i) = (n % 10) | '0';
    n /= 10;
  }
}

void l2Str(char *s, int32_t l) {
  if (l < 0L) {
    l = -l;
    s[0] = '-';
  } else s[0] = ' ';
  s[9] = '\0';  // "-12345678\0"
  for (int i = 8; i > 0; i--) {
    s[i] = (l % 10) | '0';
    l /= 10;
  }
}

int16_t l2Vstr(char *s, int32_t l) {
  uint16_t i, j = 0;
  char  temp[16];

  l2Str(temp, l);       // "-12345678\0"
  if (temp[0] == '-') s[j++] = '-';
  for (i = 1; i < 8; i++) if (temp[i] != '0') break;
  if (i > 5) {        // -678mV
    for ( ; i < 9; i++) s[j++] = temp[i];
    s[j++] = 'm';
  } else {
    if (i < 5) {      // -45.6V
      s[j++] = temp[i++];
      s[j++] = temp[i++];
      s[j++] = '.';
      s[j++] = temp[i];
    } else {        // -5.67V
      s[j++] = temp[i++];
      s[j++] = '.';
      s[j++] = temp[i++];
      s[j++] = temp[i];
    }
  }
  s[j++] = 'V';
  s[j  ] = '\0';
  return j;
}

int16_t l2Vstr5(char *s, int32_t l) {
  int16_t  i, j;
  char  temp[16];

  i = l2Vstr(temp, l);    // "-12.4V\0"
  for (i--, j = 4; i >= 0; i--) {
    if (temp[i] == '-') break;
    s[j--] = temp[i];
  }
  while (j >= 0) s[j--] = ' ';
  return ((i == 0) ? -1 : 0);
}

int16_t l2Sstr5(char *s, int32_t l) {
  int16_t  i, j;
  char  temp[16];

  i = l2Vstr(temp, l);    // "-12.4V\0"
  if (temp[i - 2] == 'm') {
    temp[i - 2] = 'u';
    temp[i - 1] = 's';
  } else
    temp[i - 1] = 'm';

  for (i--, j = 4; i >= 0; i--) {
    if (temp[i] == '-') break;
    s[j--] = temp[i];
  }
  while (j >= 0) s[j--] = ' ';
  return ((i == 0) ? -1 : 0);
}

