uint16_t tft_getFont(uint8_t *font, int16_t width, int16_t line) {
  if (width <= 8) {
    return (*(font + line) << 8);
  } else {
    if (width <= 12) {
      uint16_t n = line + (line >> 1);    // n = line * 1.5
      if (line & 1)
        return ((*(font + n) << 12) | (*(font + n + 1) << 4));
      else
        return ((*(font + n) << 8) | (*(font + n + 1) & 0xF0));
    } else {
      return (*(uint16_t *)((uint8_t *)font + ((width / 8) * line)));
    }
  }
}

void tft_drawFont(int16_t pX, int16_t pY, int16_t width, int16_t line, uint8_t *font, uint16_t fccode, uint16_t bccode) {
  uint16_t p, k;

  for (int i = 0; i < line; i++) {
    p = tft_getFont(font, width, i);
    for (int j = 0, k = 0x8000; j < width; j++, k >>= 1) {
      if (p & k) {
        M5.Lcd.drawPixel(pX + j, pY + i, Color_tbl[fccode]);
        if (drawBMP) S5drawPixel(pX + j, pY + i, fccode);
      } else if (bccode != ixBLACK1) {
        M5.Lcd.drawPixel(pX + j, pY + i, Color_tbl[bccode]);
        if (drawBMP) S5drawPixel(pX + j, pY + i, bccode);
      }
    }
  }
}

void tft_drawICON(int16_t pX, int16_t pY, int16_t width, int16_t line, uint16_t *icon, uint16_t fccode, uint16_t bccode) {
  uint16_t k;

  for (int i = 0; i < line; i++) {
    for (int j = 0, k = 0x8000; j < width; j++, k >>= 1) {
      if (*icon & k) {
        M5.Lcd.drawPixel(pX + j, pY + i, Color_tbl[fccode]);
        if (drawBMP) S5drawPixel(pX + j, pY + i, fccode);
      } else if (bccode != ixBLACK1) {
        M5.Lcd.drawPixel(pX + j, pY + i, Color_tbl[bccode]);
        if (drawBMP) S5drawPixel(pX + j, pY + i, bccode);
      }
    }
    icon++;
  }
}

int tft_drawText(int16_t pX, int16_t pY, char *s, uint16_t fccode, uint16_t bccode) {
  int l = 0;

  while (*s) {
    uint8_t cc = *s;
    if (cc > 0x8F) cc = ' ';
    tft_drawFont(pX, pY, 8, 12, &ASCIIs[cc][0], fccode, bccode);
    pX += 8;
    s++;
    l++;
  }
  return l;
}

uint16_t dodline[] = { ixGRAY2, ixBLACK, ixBLACK, ixBLACK };

void tft_drawVDotLineReps(int16_t pX, int16_t pY, int16_t dots, uint16_t fccode) {
  uint16_t wcolor = ixBLACK;

  if (fccode != ixBLACK) {
    M5.Lcd.drawFastVLine(pX, pY, dots, Color_tbl[fccode]);
    if (drawBMP) S5drawFastVLine(pX, pY, dots, fccode);
    return;
  }
  M5.Lcd.drawFastVLine(pX, pY, dots, Color_tbl[ixBLACK]);
  if (drawBMP) S5drawFastVLine(pX, pY, dots, ixBLACK);
  if (((pX - VW_xb) % 4) != 0) return;

  int Ys, Yc = dots, Yi = 20;
  if (((pX - VW_xb) % 20) == 0) Yi = 4;
  Ys = ((((pY - VW_yb + (Yi - 1)) / Yi) * Yi) - (pY - VW_yb));
  if (Ys > Yc) return;

  Yc -= Ys;
  Ys = (pY + Ys);
  wcolor = ((pX - VW_xb) == (VW_xh / 2)) ? ixYELLOW : ixGRAY2;

  for (int i = 0; i < dots; i += Yi) {
    M5.Lcd.drawPixel(pX, (Ys + i), (((Ys - VW_yb + i) == (VW_yh / 2)) ? Color_tbl[ixYELLOW] : Color_tbl[wcolor]));
    if (drawBMP) S5drawPixel(pX, (Ys + i), (((Ys - VW_yb + i) == (VW_yh / 2)) ? ixYELLOW : wcolor));
  }
}

void tft_drawVDotLineRep(int16_t pX, int16_t pY1, int16_t pY2, uint16_t fccode) {
  if (pY1 < pY2)
    tft_drawVDotLineReps(pX, pY1, (pY2 - pY1 + 1), fccode);
  else
    tft_drawVDotLineReps(pX, pY2, (pY1 - pY2 + 1), fccode);
}


uint16_t dodlineG[] = { ixWHITE, ixBLACK, ixBLACK, ixBLACK };

void tft_drawHDotLine(int16_t pX, int16_t pY, int16_t dots, uint16_t fccode) {
  for (int i = 0; i < dots; i++) {
    M5.Lcd.drawPixel(pX + i, pY, ((dodlineG[i & 3] == ixBLACK) ? Color_tbl[ixBLACK] : Color_tbl[fccode]));
    if (drawBMP) S5drawPixel(pX + i, pY, ((dodlineG[i & 3] == ixBLACK) ? ixBLACK : fccode));
  }
}

void tft_drawVDotLine(int16_t pX, int16_t pY, int16_t dots, uint16_t fccode) {
  for (int i = 0; i < dots; i++) {
    M5.Lcd.drawPixel(pX, pY + i, ((dodlineG[i & 3] == BLACK) ? Color_tbl[ixBLACK] : Color_tbl[fccode]));
    if (drawBMP) S5drawPixel(pX, pY + i, ((dodlineG[i & 3] == BLACK) ? ixBLACK : fccode));
  }
}

void tft_fillRect(int16_t pX, int16_t pY, int16_t w, int16_t h, uint16_t fccode) {
  M5.Lcd.fillRect(pX, pY, w, h, Color_tbl[fccode]);
  if (drawBMP) S5fillRect(pX, pY, w, h, fccode);
}

void tft_drawRect(int16_t pX, int16_t pY, int16_t w, int16_t h, uint16_t fccode) {
  M5.Lcd.drawRect(pX, pY, w, h, Color_tbl[fccode]);
  if (drawBMP) S5drawRect(pX, pY, w, h, fccode);
}

void tft_drawFastHLine(int16_t pX, int16_t pY, int16_t w, uint16_t fccode) {
  M5.Lcd.drawFastHLine(pX, pY, w, Color_tbl[fccode]);
  if (drawBMP) S5drawFastHLine(pX, pY, w, fccode);
}

void S5drawPixel(int16_t pX, int16_t pY, uint16_t fccode) {
  if (pX < 0 || pX >= 320 || pY < 0 || pY >= 240) Serial.printf("%s = %d, %d\n", __func__, pX, pY);
  bmpBuff[pY][pX] = fccode;
}

void S5drawFastHLine(int16_t pX, int16_t pY, int16_t w, uint16_t fccode) {
  if (pX < 0 || pX >= 320 || pY < 0 || pY >= 240 || (pX + w) > 320) Serial.printf("%s - %d, %d, %d\n", __func__, pX, pY, pX + w);
  for (int i = 0; i < w; i++)
    bmpBuff[pY][pX + i] = fccode;
}

void S5drawFastVLine(int16_t pX, int16_t pY, int16_t h, uint16_t fccode) {
  if (pX < 0 || pX >= 320 || pY < 0 || pY >= 240 || (pY + h) > 240) Serial.printf("%s - %d, %d, %d\n", __func__ , pX, pY, pY + h);
  for (int i = 0; i < h; i++)
    bmpBuff[pY + i][pX] = fccode;
}

void S5drawRect(int16_t pX, int16_t pY, int16_t w, int16_t h, uint16_t fccode) {
  if (pX < 0 || pX >= 320 || pY < 0 || pY >= 240 || (pX + w) > 320 || (pY + h) > 240) Serial.printf("%s - %d, %d, %d, %d\n", __func__, pX, pY, pX + w, pY + h);
  S5drawFastHLine(pX,        pY,         w, fccode);
  S5drawFastVLine(pX,        pY,         h, fccode);
  S5drawFastHLine(pX,        pY + h - 1,  w, fccode);
  S5drawFastVLine(pX + w - 1, pY,         h, fccode);
}

void S5fillRect(int16_t pX, int16_t pY, int16_t w, int16_t h, uint16_t fccode) {
  if (pX < 0 || pX >= 320 || pY < 0 || pY >= 240 || (pX + w) > 320 || (pY + h) > 240) Serial.printf("%s - %d, %d, %d, %d\n", __func__, pX, pY, pX + w, pY + h);
  for (int i = 0; i < h; i++)
    S5drawFastHLine(pX, pY + i, w, fccode);
}

void drawBar(int pos) {
  tft_fillRect(50 + (pos * 2), 102, 2, 2, ixWHITE);
}

void dumpScreen(int out) {
  char cbuff[64];

  drawBMP = 0;
  if (out == SV_DP_SEL_BMP) dumpFilename(cbuff); else sprintf(cbuff, "Dump to UART");

  tft_fillRect(30,  70, 240, 50, ixBLUE);
  tft_drawRect(29,  69, 242, 52, ixWHITE);
  tft_drawText(50,  80, cbuff, ixWHITE, ixBLUE);
  tft_fillRect(50, 102, 200,  2, ixGRAY1);

  if (out == SV_DP_SEL_BMP) {
    int rc = dumpx(sv_dp_sel, cbuff, drawBar);
    if (rc == 0) sprintf(cbuff, "Saved             ");
    else sprintf(cbuff, "Save error (%d)    ", rc);
    tft_drawText( 50,  96, cbuff, ixWHITE, ixBLUE);
    tft_drawText(190,  96, "Press[A]", ixYELLOW, ixBLUE);
    while (!M5.BtnA.wasPressed())  {
      M5.update(); delay(20);
    }
    M5.update();
  } else {
    dumpUART();
  }
  drawBMP = 1;

  for (size_t y = 0; y < TFT_MAX_Y; y++)
    for (size_t x = 0; x < TFT_MAX_X; x++)
      M5.Lcd.drawPixel(x, y, Color_tbl[bmpBuff[y][x]]);
}

void dumpUART() {
  for (int i = 0; i < VW_xh; i++) {
    Serial.printf("%d, ", (sv_ch_sel & SV_CH1) ? ((VW_yb + VW_yh) - waverec[0][1][i]) : 0);
    Serial.printf("%d\n", (sv_ch_sel & SV_CH2) ? ((VW_yb + VW_yh) - waverec[1][1][i]) : 0);
    drawBar((i * 100) / VW_xh);
  }
}

